import os
from functools import wraps
from pathlib import Path
from urllib.parse import unquote

from fastapi import APIRouter, Body, Form
from fastapi.params import Depends
from loguru import logger
from starlette.requests import Request
from starlette.responses import HTMLResponse, JSONResponse

from shipaw.fapi.alerts import Alerts
from shipaw.fapi.emailer import send_label_email
from shipaw.fapi.form_data import shipment_request_form, shipment_request_form_json
from shipaw.fapi.requests import ShipmentRequest
from shipaw.fapi.responses import ShipawTemplateResponse
from shipaw.fapi.routes_api import (
    order_results_api as order_confirm_json,
    order_summary_api as order_review_json,
    shipping_form_api as ship_form_json,
)
from shipaw.models.shipment import Shipment, sample_shipment

router = APIRouter()


def render_template_response(request: Request, resp: ShipawTemplateResponse) -> HTMLResponse:
    context = resp.template.context
    context['alerts'] = context.get('alerts', Alerts.empty()) + resp.alerts
    if resp.alerts.errors:
        return request.app.shipaw_settings.templates.TemplateResponse(
            request=request,
            name='alerts.html',
            context=context,
        )

    return request.app.shipaw_settings.templates.TemplateResponse(
        request=request,
        name=resp.template.template_path,
        context=context,
    )


def html_from_json(json_endpoint):
    @wraps(json_endpoint)
    async def wrapper(request: Request, *args, **kwargs):
        resp = await json_endpoint(request, *args, **kwargs)

        return request.app.shipaw_settings.templates.TemplateResponse(
            request=request, name=resp.template_path, context=resp.context
        )

    return wrapper


@router.post('/shipping_form', response_class=HTMLResponse)
async def shipping_form(request: Request, shipment: Shipment = Body(...)) -> HTMLResponse:
    res = await ship_form_json(request, shipment)
    return render_template_response(request, res)


@router.post('/order_summary', response_class=HTMLResponse)
async def order_summary(
    request: Request,
    shipment_request: ShipmentRequest = Depends(shipment_request_form),
) -> HTMLResponse:
    res = await order_review_json(request, shipment_request)
    return render_template_response(request, res)


@router.post('/order_results', response_class=HTMLResponse)
async def order_results(
    request: Request,
    shipment_request: ShipmentRequest = Depends(shipment_request_form_json),
) -> HTMLResponse:
    template_response = await order_confirm_json(request, shipment_request)
    return render_template_response(request, template_response)


@router.get('/home_mobile_phone', response_class=HTMLResponse)
async def home_mobile_phone(request: Request) -> HTMLResponse:
    mobile_phone = request.app.shipaw_settings.mobile_phone
    return f"""
    <input type="tel" id="mobile_phone" name="mobile_phone" value="{mobile_phone}" required>
    """


@router.get('/open-file/{filepath}', response_class=HTMLResponse)
async def open_file(filepath: str):
    os.startfile(filepath)
    return HTMLResponse(content='<span>Re</span>')


@router.get('/print-file/{filepath}', response_class=HTMLResponse)
async def print_file(filepath: str):
    os.startfile(filepath, 'print')
    return HTMLResponse(content='<span>Re</span>')


@router.post('/email-label', response_class=HTMLResponse)
async def email_label(
    label_path: str = Form(...), shipment_request: ShipmentRequest = Depends(shipment_request_form_json)
):
    label_path = Path(unquote(label_path))
    logger.info(f'Emailing {label_path=} to {shipment_request.shipment.remote_full_contact.contact.email_address}')
    await send_label_email(shipment_request, label_path)
    return HTMLResponse(content='<span>Re</span>')


@router.get('/health', response_class=JSONResponse)
async def health(request: Request) -> JSONResponse:
    return JSONResponse({'status': 'ok'})


@router.get('/test', response_class=HTMLResponse)
async def test_route(request: Request) -> HTMLResponse:
    shipment = sample_shipment()
    res = await ship_form_json(request, shipment)
    return render_template_response(request, res)


@router.get('/', response_class=HTMLResponse)
async def test_route2(request: Request) -> HTMLResponse:
    shipment = sample_shipment()
    res = await ship_form_json(request, shipment)
    return render_template_response(request, res)



import requests
import webbrowser
import tempfile


def do_track(consignment_number: int, postcode: str):
    url = 'https://apcchoice.apc-overnight.com/APCChoice'
    form_data = {'ConsignmentNumber': consignment_number, 'Postcode': postcode}

    response = requests.post(url, data=form_data)

    with tempfile.NamedTemporaryFile(delete=False, suffix='.html') as temp_file:
        temp_file.write(response.content)
        temp_file_path = temp_file.name

    webbrowser.open(f'file://{temp_file_path}')


if __name__ == '__main__':
    do_track(39, 'TF1 2NP')



from typing import TYPE_CHECKING

from shipaw.models.base import ShipawBaseModel

if TYPE_CHECKING:
    ...
from shipaw.fapi.requests import ShipmentRequest
from shipaw.fapi.responses import ShipmentResponse


class ShipmentBooking(ShipawBaseModel):
    request: 'ShipmentRequest'
    response: 'ShipmentResponse'



from shipaw.providers.apc import APCShippingProvider, APCSettings
from shipaw.providers.royal_mail import RoyalMailProvider, RoyalMailSettingsGlobal

__all__ = [
    'APCSettings',
    'APCShippingProvider',
    'RoyalMailProvider',
    'RoyalMailSettingsGlobal',
]



from apc_hypaship.models.request.address import Address, Contact

from shipaw.models.address import Address as AddressAgnost, Contact as ContactAgnost, FullContact


def address_from_agnostic_fc[addr_type: Address](cls: type[addr_type], full_contact: FullContact) -> addr_type:
    lines_ = [_ for _ in full_contact.address.address_lines[1:] if _]
    lines = ', '.join(lines_)
    return cls(
        company_name=full_contact.address.business_name,
        address_line_1=full_contact.address.address_lines[0],
        address_line_2=lines,
        city=full_contact.address.town,
        postal_code=full_contact.address.postcode,
        country_code=full_contact.address.country,
        contact=Contact(
            person_name=full_contact.contact.contact_name,
            email=full_contact.contact.email_address,
            mobile_number=full_contact.contact.mobile_phone,
            phone_number=full_contact.contact.phone_number or full_contact.contact.mobile_phone,
        ),
    )


def contact_from_agnostic_fc[contact_type: Contact](cls: type[contact_type], full_contact: FullContact) -> contact_type:
    return cls(
        person_name=full_contact.contact.contact_name,
        email=full_contact.contact.email_address,
        mobile_number=full_contact.contact.mobile_phone,
        phone_number=full_contact.contact.phone_number,
    )


def full_contact_from_apc_contact_address(contact: Contact, address: Address) -> FullContact:
    return FullContact(
        address=AddressAgnost(
            business_name=address.company_name,
            address_lines=[line for line in [address.address_line_1, address.address_line_2] if line],
            town=address.city,
            postcode=address.postal_code,
            country=address.country_code,
        ),
        contact=ContactAgnost(
            contact_name=contact.person_name,
            email_address=contact.email,
            mobile_phone=contact.mobile_number,
            phone_number=contact.phone_number or contact.mobile_number,
        ),
    )



import time
from abc import ABC, abstractmethod
from enum import StrEnum
from typing import ClassVar, TYPE_CHECKING

from pydantic import BaseModel, Field
from pydantic_settings import BaseSettings

from shipaw.models.base import ShipawBaseModel
from shipaw.models.ship_types import ShipDirection
from shipaw.models.shipment import Shipment

if TYPE_CHECKING:
    from shipaw.fapi.responses import ShipmentResponse
    from shipaw.fapi.requests import ShipmentRequest


class ProviderName(StrEnum):
    ROYAL_MAIL = 'ROYAL_MAIL'
    APC = 'APC'


class HasServiceCodes(ABC):
    service_codes_type: ClassVar[type[StrEnum]]
    default_service: ClassVar[StrEnum]

    @classmethod
    def lookup_service(cls, service_name: str):
        return cls.service_codes_type[service_name]

    @classmethod
    def reverse_lookup_service(cls, service_code: str):
        return cls.service_codes_type(service_code).name

    @classmethod
    def services_as_tups(cls) -> list[tuple[str, str]]:
        return [(e.name, str(e.value)) for e in cls.service_codes_type]

    @classmethod
    def services_as_dict(cls) -> dict[str, str]:
        return dict(cls.service_codes_type.__members__.items())


class HasLabels(ABC):
    @abstractmethod
    def fetch_label_content(self, shipment_num: str) -> bytes: ...

    def wait_fetch_label(self, shipment_num: str, tries=10) -> bytes:
        for i in range(tries):
            try:
                time.sleep(1)  # let API process booking
                label_data = self.fetch_label_content(shipment_num=shipment_num)
                assert label_data is not None, f'Label not ready yet for {shipment_num}, retrying...'
                return label_data
            except AssertionError:
                pass
        raise RuntimeError(f'Label not ready after {tries} retries for {shipment_num}')

    async def wait_fetch_label_async(self, shipment_num: str, tries=10) -> bytes:
        for i in range(tries):
            try:
                time.sleep(1)
                label_data = self.fetch_label_content(shipment_num=shipment_num)
                assert label_data is not None
                return label_data
            except AssertionError:
                print(f'Label not ready yet for {shipment_num}, retrying...')
        raise RuntimeError(f'Label not ready after {tries}retries for {shipment_num}')


class ShippingProvider(HasServiceCodes, HasLabels, ABC, ShipawBaseModel):
    name: ClassVar[ProviderName]
    settings: BaseSettings
    settings_type: ClassVar[type[BaseSettings]]
    valid_directions: ClassVar[list[ShipDirection]]

    valid_direction_services: ClassVar[dict[ShipDirection, list[StrEnum]]] = Field(default_factory=dict)

    @property
    @abstractmethod
    def client(self):
        raise NotImplementedError

    @abstractmethod
    def is_sandbox(self) -> bool: ...

    @staticmethod
    @abstractmethod
    def provider_shipment(shipment: Shipment, service_code: StrEnum) -> BaseModel:
        """Takes Shipaw's agnostic Shipment object and returns provider-specific Shipment object"""
        ...

    @staticmethod
    @abstractmethod
    def agnostic_shipment(shipment: BaseModel) -> Shipment:
        """Takes provider-specific Shipment object and returns Shipaw's agnostic Shipment object"""
        ...

    @staticmethod
    @abstractmethod
    def book_shipment_agnostic(shipment_request: 'ShipmentRequest') -> 'ShipmentResponse': ...



from base64 import b64decode
from typing import ClassVar, override

from apc_hypaship.apc_client import APCClient
from apc_hypaship.config import APCSettings
from apc_hypaship.models.request.address import Address
from apc_hypaship.models.request.services import APCServiceCode
from apc_hypaship.models.request.shipment import GoodsInfo, Order, Orders, Shipment as ShipmentAPC, ShipmentDetails
from apc_hypaship.models.response.common import APCException
from apc_hypaship.models.response.resp import BookingResponse

from shipaw.providers.registry import register_provider_type
from shipaw.fapi.requests import ShipmentRequest
from shipaw.fapi.responses import ShipmentResponse
from shipaw.logging import log_obj
from shipaw.models.ship_types import ShipDirection
from shipaw.models.shipment import Shipment as ShipmentAgnost
from shipaw.providers.apc.apc_funcs import (
    address_from_agnostic_fc,
    full_contact_from_apc_contact_address,
)
from shipaw.providers.apc.response import errored_booking
from shipaw.providers.provider_abc import ProviderName, ShippingProvider


@register_provider_type
class APCShippingProvider(ShippingProvider):
    settings: APCSettings

    name: ClassVar[ProviderName] = ProviderName.APC

    settings_type: ClassVar[type[APCSettings]] = APCSettings
    service_codes_type: ClassVar[type[APCServiceCode]] = APCServiceCode
    default_service: ClassVar[APCServiceCode] = APCServiceCode.PARCEL_1600
    _client: APCClient | None = None

    valid_directions: ClassVar[list[ShipDirection]] = [ShipDirection.OUTBOUND, ShipDirection.INBOUND]

    valid_direction_services = {
        ShipDirection.OUTBOUND: [APCServiceCode.PARCEL_1600, APCServiceCode.SATURDAY_PARCEL_1200],
        ShipDirection.INBOUND: [APCServiceCode.PARCEL_1600],
    }

    @override
    def is_sandbox(self) -> bool:
        return 'training' in self.settings.base_url.lower()

    @property
    def client(self) -> APCClient:
        if self._client is None:
            if self.settings is None:
                raise ValueError('Settings must be set before using the client')
            self._client = APCClient(settings=self.settings)
        return self._client

    @override
    def provider_shipment(self, shipment: ShipmentAgnost, service_code: APCServiceCode) -> ShipmentAPC:
        if shipment.direction == ShipDirection.DROPOFF:
            raise NotImplementedError('APC Do not do dropoffs')
        order = Order(
            ready_at=shipment.collect_ready,
            closed_at=shipment.collect_closed,
            collection_date=shipment.shipping_date,
            product_code=service_code,
            reference=shipment.reference,
            delivery=address_from_agnostic_fc(Address, shipment.recipient),
            collection=address_from_agnostic_fc(Address, shipment.sender) if shipment.sender else None,
            goods_info=GoodsInfo(),
            shipment_details=ShipmentDetails(number_of_pieces=shipment.boxes),
        )
        return ShipmentAPC(orders=Orders(order=order))

    @override
    def agnostic_shipment(self, shipment: ShipmentAPC) -> ShipmentAgnost:
        """Takes APC Shipment object, returns agnostic Shipment object"""
        order = shipment.orders.order
        del_fc = full_contact_from_apc_contact_address(order.delivery.contact, order.delivery)
        send_fc = (
            full_contact_from_apc_contact_address(order.collection.contact, order.collection)
            if order.collection
            else None
        )
        return ShipmentAgnost(
            shipping_date=order.collection_date,
            reference=order.reference,
            recipient=del_fc,
            sender=send_fc,
            boxes=order.shipment_details.number_of_pieces,
            direction=ShipDirection.INBOUND if order.collection is not None else ShipDirection.OUTBOUND,
            collect_ready=order.ready_at,
            collect_closed=order.closed_at,
        )

    @override
    def book_shipment_agnostic(self, shipment_request: ShipmentRequest) -> 'ShipmentResponse':
        """Takes provider ShipmnentDict, or ShipmentAgnost object"""
        # request_json = self.build_request_json(shipment)
        provider_service = self.service_codes_type(shipment_request.service_code)
        shipment = shipment_request.shipment
        provider_shipment = self.provider_shipment(shipment, provider_service)
        log_obj(provider_shipment, 'APC Shipment Request')
        try:
            apc_response: BookingResponse = self.client.fetch_book_shipment(provider_shipment)
        except APCException as e:
            return errored_booking(shipment, e)  # should be a response not exception?
        response = self.build_response(apc_response, shipment)
        response.label_data = self.wait_fetch_label(response.shipment_num)
        return response

    @override
    def fetch_label_content(self, shipment_num: str) -> bytes:
        labl = self.client.fetch_label(shipment_num)
        content = labl.content
        return b64decode(content)

    @staticmethod
    def build_response(resp: BookingResponse, shipment: ShipmentAgnost):
        orders = resp.orders
        order = orders.order
        return ShipmentResponse(
            shipment=shipment,
            shipment_num=order.order_number,
            tracking_link=r'https://apc.hypaship.com/app/shared/customerordersoverview/index#search_form',
            data=resp.model_dump(mode='json'),
            status=(str(orders.messages.code)),
            success=(orders.messages.code == 'SUCCESS'),
        )



import contextlib
import os

from fastapi import FastAPI, Query, responses
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from loguru import logger
from pawlogger import configure_loguru
from starlette.requests import Request
from starlette.responses import HTMLResponse, JSONResponse
from starlette.staticfiles import StaticFiles

from shipaw.fapi.alerts import Alert, AlertType, Alerts
from shipaw.fapi.routes_api import router as json_router
from shipaw.fapi.routes_html import router as html_router
from shipaw.config import SHIPAW_SETTINGS, populate_providers


@contextlib.asynccontextmanager
async def lifespan(app_: FastAPI):
    try:
        app_.shipaw_settings = SHIPAW_SETTINGS
        log_file = SHIPAW_SETTINGS.log_file
        configure_loguru(logger, log_file=log_file, level=SHIPAW_SETTINGS.log_level)
        populate_providers(SHIPAW_SETTINGS)
        yield

    finally:
        # pythoncom.CoUninitialize()

        ...


app = FastAPI(lifespan=lifespan)
app.mount('/static', StaticFiles(directory=str(SHIPAW_SETTINGS.static_dir)), name='static')
app.include_router(json_router, prefix='/api')
app.include_router(html_router)
app.ship_live = SHIPAW_SETTINGS.shipper_live
app.alerts = Alerts.empty()


@app.exception_handler(RequestValidationError)
async def request_validation_exception_handler(request: Request, exc: RequestValidationError):
    errors = exc.errors()
    msg2 = ''
    for err in errors:
        if err_type := err.get('type'):
            msg2 += err_type + ' '
        if loc := err.get('loc'):
            msg2 += f'in {loc} '
        if ctx := err.get('ctx'):
            if reason := ctx.get('reason'):
                msg2 += f': {reason} '
        if input_ := err.get('input'):
            msg2 += f'. Input = {input_} '

    logger.error(msg2)
    alerts = Alerts(alert=[Alert(code=1, message=msg2, type=AlertType.ERROR)])
    return JSONResponse(
        status_code=422,
        content={'detail': jsonable_encoder(exc.errors()), 'alerts': alerts.model_dump(mode='json'), 'message': msg2},
    )


@app.get('/robots.txt', response_class=responses.PlainTextResponse)
async def robots_txt() -> str:
    return 'User-agent: *\nAllow: /'


@app.get('/favicon.ico', include_in_schema=False)
async def favicon_ico():
    return responses.RedirectResponse(url='/static/favicon.svg')


@app.get('/', response_class=JSONResponse)
async def base():
    return JSONResponse(content={'status': 'ok'})


@app.get('/open-file', response_class=HTMLResponse)
async def open_file(filepath: str = Query(...)):
    os.startfile(filepath)
    return HTMLResponse(content='<span>Re</span>')
